/*
 * This file is part of xsoundtrack - a sound tracker for Linux.
 * Sed - july august 1998.
 * This is total free software.
 *-------------------------------------
 * a simple file selector
 */

#include "file_selector.h"

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <dirent.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>

#include "interface.h"
#include "screen.h"

/* the height of the window */
#define HAUTEUR (38+14*16+20+4+4+20+4)

/* the file/directorie-oriented functions */

/* a sort function for qsort. Used by get_directories_files */
int comp(char **a, char **b)
{
  return strcmp(*a, *b);
}

void reset_file_dir(INTERFACE *f)
{
  int i;
  for (i=0; i<f->file_selector.nb_dirs; i++)
    if ((f->file_selector.dirs)[i])
      free((f->file_selector.dirs)[i]);
  if (f->file_selector.dirs)
    free(f->file_selector.dirs);
  f->file_selector.dirs=(char **)0;
  f->file_selector.nb_dirs=0;

  for (i=0; i<f->file_selector.nb_files; i++)
    if ((f->file_selector.files)[i])
      free((f->file_selector.files)[i]);
  if (f->file_selector.files)
    free(f->file_selector.files);
  f->file_selector.files=(char **)0;
  f->file_selector.nb_files=0;
}

void get_directories_files(char ***file_return, char ***dir_return, int *nb_file_return, int *nb_dir_return)
/* d'you like *** ? hehe, absolutly useless, i simply could give INTERFACE *f and it would be the same */
/* but i like it */
/* and change it now... you know... i surely would introduce bugs here and there... */
{
  DIR *dir;
  struct dirent *dirent;

  *file_return=(char **)0;
  *dir_return=(char **)0;
  *nb_file_return=0;
  *nb_dir_return=0;

  /* dir exists ? */
  if (!(dir=opendir("."))) {
    perror("file_selector.get_directories_files error 0\nperror says");
    chdir("/");
    if (!(dir=opendir(".")))
      return;
  }

  while((dirent=readdir(dir))) {
    struct stat the_stat;
    char **dum;
    /* test if it is a directory */
    if (lstat(dirent->d_name, &the_stat)==-1) {
      fprintf(stderr, "dirent->d_name=\"%s\"", dirent->d_name);
      perror("file_selector.get_directories_files error 3\nperror says");
      goto error;
    }
    if (S_ISDIR(the_stat.st_mode)) {
      /* it's a dir */
      /* we don't want "." */
      if (!strcmp(dirent->d_name, "."))
	continue;
      dum=(char **)realloc(*dir_return, ((*nb_dir_return)+1)*sizeof(char *));
      if (!dum)
	goto error;
      *dir_return=dum;
      if (!((*dir_return)[(*nb_dir_return)++]=strdup(dirent->d_name)))
	goto error;
      continue;
    }
    /* it's not a dir, consider it as a file (it may be a device, or a symlink, or anything else than a dir) */
    dum=(char **)realloc(*file_return, ((*nb_file_return)+1)*sizeof(char *));
    if (!dum)
      goto error;
    *file_return=dum;
    if (!((*file_return)[(*nb_file_return)++]=strdup(dirent->d_name)))
      goto error;
  }

  /* now, sort all with qsort */
  if (*nb_dir_return)
    qsort(*dir_return, *nb_dir_return, sizeof(char *), (int(*)(const void *, const void *))&comp);
  if (*nb_file_return)
    qsort(*file_return, *nb_file_return, sizeof(char *), (int(*)(const void *, const void *))&comp);

  goto ok;

error: {
    int i;
    for (i=0; i<*nb_dir_return; i++)
      if ((*dir_return)[i])
	free((*dir_return)[i]);
    if (*dir_return)
      free(*dir_return);
    *dir_return=(char **)0;
    *nb_dir_return=0;

    for (i=0; i<*nb_file_return; i++)
      if ((*file_return)[i])
	free((*file_return)[i]);
    if (*file_return)
      free(*file_return);
    *file_return=(char **)0;
    *nb_file_return=0;
    chdir("/");
  }
ok:
  closedir(dir);
  return;
}

/* the interface oriented functions */

static int around[5][4]={
  {4, 4, 500-8, 30},
  {4, 38, (500-8)/2-2, 14*16+4+20},
  {4+(500-8)/2+2, 38, (500-8)/2-2, 14*16+4+20},
  {250-70, 38+14*16+4+20+4, 60, 20},
  {250+10, 38+14*16+4+20+4, 60, 20},
};

static unsigned long color[5]={
  GREY,
  GREY,
  GREY,
  LIGHTBLUE,
  LIGHTBLUE
};

void around_file_selector(INTERFACE *i, int n, int type)
{
  draw_rectangle(i->X_ecran, i->file_selector.pixmap, i->context, around[n][0]-2, around[n][1]-2,
		 around[n][2]+4, around[n][3]+4, 2, type == IN ? WHITE : BLACK);
  draw_rectangle(i->X_ecran, i->file_selector.window, i->context, around[n][0]-2, around[n][1]-2,
		 around[n][2]+4, around[n][3]+4, 2, type == IN ? WHITE : BLACK);
}

void _draw_out_name(INTERFACE *f)
{
  draw_box(f->X_ecran, f->file_selector.pixmap, f->context, 10, 28-10-1, 500-8-6-4, 13, LIGHTBLUE);
  if (f->file_selector.out_name) {
    int len=strlen(f->file_selector.out_name);
    if (len>80) {
      draw_string(f->X_ecran, f->file_selector.pixmap, f->context, 12+3*6, 28, &f->file_selector.out_name[len-80+3],
		  f->little, BLACK);
      draw_string(f->X_ecran, f->file_selector.pixmap, f->context, 12, 28, "...", f->little, RED);
    } else
      draw_string(f->X_ecran, f->file_selector.pixmap, f->context, 12, 28, f->file_selector.out_name, f->little, BLACK);
  }
}

void draw_out_name(INTERFACE *f)
{
  _draw_out_name(f);
  XCopyArea(f->X_ecran, f->file_selector.pixmap, f->file_selector.window, f->context,
	    10, 28-10-1, 500-8-6-4, 13, 10, 28-10-1);
}

void _draw_dir(INTERFACE *f)
{
  int i, len;

  draw_box(f->X_ecran, f->file_selector.pixmap, f->context, 8, 70-14+1, (500-8)/2-2-8, 14*16+1, LIGHTBLUE);

  if (!f->file_selector.nb_dirs)
    draw_string(f->X_ecran, f->file_selector.pixmap, f->context, 20, 165, "-=[ No sub-directory ]=-", f->medium, BLACK);
  else {
    draw_box(f->X_ecran, f->file_selector.pixmap, f->context, 8+1, 1+70-14+1+14*f->file_selector.dir_pos,
	     (500-8)/2-2-8-2, 13, WHITE);
    for (i=0; i<16; i++) {
      if (i+f->file_selector.dir_start>=f->file_selector.nb_dirs)
	break;
      len=strlen(f->file_selector.dirs[i+f->file_selector.dir_start]);
      #define s ((f->file_selector.dirs[i+f->file_selector.dir_start])+(len>39 ? len : 39)-39)
      if (len<=39)
	draw_string(f->X_ecran, f->file_selector.pixmap, f->context, 10, 68+i*14, s, f->little, BLUE);
      else {
	draw_string(f->X_ecran, f->file_selector.pixmap, f->context, 10+3*6, 68+i*14, &s[3], f->little, BLUE);
	draw_string(f->X_ecran, f->file_selector.pixmap, f->context, 10, 68+i*14, "...", f->little, RED);
      }
    }
  }
}

void _draw_file(INTERFACE *f)
{
  int i, len;

  draw_box(f->X_ecran, f->file_selector.pixmap, f->context, 4+(500-8)/2+2+4, 70-14+1, (500-8)/2-2-8, 14*16+1, LIGHTBLUE);

  if (!f->file_selector.nb_files)
    draw_string(f->X_ecran, f->file_selector.pixmap, f->context, 59+492/2+8, 165, "-=[ No file ]=-", f->medium, BLACK);
  else {
    draw_box(f->X_ecran, f->file_selector.pixmap, f->context, 1+4+(500-8)/2+2+4, 1+70-14+1+14*f->file_selector.file_pos,
	     (500-8)/2-2-8-2, 13, WHITE);
    for (i=0; i<16; i++) {
      if (i+f->file_selector.file_start>=f->file_selector.nb_files)
	break;
      len=strlen(f->file_selector.files[i+f->file_selector.file_start]);
      #undef s
      #define s ((f->file_selector.files[i+f->file_selector.file_start])+(len>39 ? len : 39)-39)
      if (len<=39)
	draw_string(f->X_ecran, f->file_selector.pixmap, f->context, 4+(500-8)/2+2+6, 68+i*14, s, f->little, BLUE);
      else {
	draw_string(f->X_ecran, f->file_selector.pixmap, f->context, 4+(500-8)/2+2+6+3*6, 68+i*14, &s[3], f->little, BLUE);
	draw_string(f->X_ecran, f->file_selector.pixmap, f->context, 4+(500-8)/2+2+6, 68+i*14, "...", f->little, RED);
      }
      #undef s
    }
  }
}

void _draw_dir_and_file_names(INTERFACE *f)
{
  if (f->file_selector.out_name)
    draw_string(f->X_ecran, f->file_selector.pixmap, f->context, 12, 27, f->file_selector.out_name, f->little, BLUE);

  _draw_dir(f);
  _draw_file(f);
}

void draw_dir(INTERFACE *f)
{
  _draw_dir(f);
  XCopyArea(f->X_ecran, f->file_selector.pixmap, f->file_selector.window, f->context,
	    8, 70-14+1, (500-8)/2-2-8, 14*16+1, 8, 70-14+1);
}

void draw_file(INTERFACE *f)
{
  _draw_file(f);
  XCopyArea(f->X_ecran, f->file_selector.pixmap, f->file_selector.window, f->context,
	    4+(500-8)/2+2+4, 70-14+1, (500-8)/2-2-8, 14*16+1, 4+(500-8)/2+2+4, 70-14+1);
}

void draw_dir_and_file_names(INTERFACE *f)
{
  draw_dir(f);
  draw_file(f);
}

void draw_name_cursor(INTERFACE *f, int type)
{
  int len=0;
  if (f->file_selector.out_name) {
    len=strlen(f->file_selector.out_name);
    if (len>80)
      len=80;
  }

  draw_line(f->X_ecran, f->file_selector.pixmap, f->context, 12+6*len, 18, 12+6*len, 29, type==IN ? BLACK : LIGHTBLUE);
  draw_line(f->X_ecran, f->file_selector.window, f->context, 12+6*len, 18, 12+6*len, 29, type==IN ? BLACK : LIGHTBLUE);
}

void redo_around(INTERFACE *f, int n, unsigned long mode)
{
  draw_3d_box_around(f->X_ecran, f->file_selector.pixmap, f->context, around[n][0], around[n][1],
		     around[n][2], around[n][3], color[n], mode);
  draw_3d_box_around(f->X_ecran, f->file_selector.window, f->context, around[n][0], around[n][1],
		     around[n][2], around[n][3], color[n], mode);
}

void enter_win_sel_0(INTERFACE *f)
{
  redo_around(f, 0, DOWN_3D);
  draw_name_cursor(f, IN);
}

void enter_win_sel_1(INTERFACE *f)
{
  redo_around(f, 1, DOWN_3D);
}

void enter_win_sel_2(INTERFACE *f)
{
  redo_around(f, 2, DOWN_3D);
}

void enter_win_sel_3(INTERFACE *f)
{
  redo_around(f, 3, DOWN_3D);
  around_file_selector(f, 3, OUT);
  XFlush(f->X_ecran);		/* this one is needed */
  f->file_selector.in_use=0;
}

void enter_win_sel_4(INTERFACE *f)
{
  redo_around(f, 4, DOWN_3D);
  if (f->file_selector.out_name) {
    free(f->file_selector.out_name);
    f->file_selector.out_name=(char *)0;
  }
  f->file_selector.in_use=0;
}

void exit_win_sel_0(INTERFACE *f)
{
  redo_around(f, 0, UP_3D);
  draw_name_cursor(f, OUT);
}

void exit_win_sel_1(INTERFACE *f)
{
  redo_around(f, 1, UP_3D);
}

void exit_win_sel_2(INTERFACE *f)
{
  redo_around(f, 2, UP_3D);
}

void interface_win_sel_0(INTERFACE *f, KeySym ks)
{
  char *tmp;
  int len;

  if (ks==XK_BackSpace) {
    if (!f->file_selector.out_name) {
      XBell(f->X_ecran, 0);
      return;
    }
    len=strlen(f->file_selector.out_name);
    if (len==0) {
      /* should not be possible */
      fprintf(stderr, "file_selector.interface_win_sel_0 impossible len == 0\n");
      return;
    }
    /* no test should be necessary, because we reduce the size of the area. Well, bad code anyway. */
    f->file_selector.out_name=realloc(f->file_selector.out_name, len);
    f->file_selector.out_name[--len]=(char)0;
    if (len==0) {
      free(f->file_selector.out_name);
      f->file_selector.out_name=(char *)0;
    }
    draw_out_name(f);
    draw_name_cursor(f, IN);
  }

  /* only latin-1 now */
  if ((unsigned long)ks<XK_space || (unsigned long)ks>XK_ydiaeresis)
    return;

  /* ok, so now put it */
  len=0;
  if (f->file_selector.out_name) {
    len=strlen(f->file_selector.out_name);
    /* should not be 0-length */
    if (len==0) {
      fprintf(stderr, "file_selector.interface_win_sel_0 impossible 2 len == 0\n");
      return;
    }
  }
  tmp=realloc(f->file_selector.out_name, (++len)+1);
  if (!tmp) {
    fprintf(stderr, "file_selector.interface_win_sel_0 error realloc failed\n");
    return;
  }
  f->file_selector.out_name=tmp;
  tmp[len-1]=(char)ks;
  tmp[len]=(char)0;
  draw_out_name(f);
  draw_name_cursor(f, IN);
}

void interface_win_sel_1(INTERFACE *f, KeySym ks)
{
  switch(ks) {
    case XK_Down :
      /* this test is enough, it can't be higher (0>-1), because if nb_files==0, we never will be here */
      if (f->file_selector.dir_start+f->file_selector.dir_pos==f->file_selector.nb_dirs-1)
	break;
      if (f->file_selector.dir_pos==15)
	f->file_selector.dir_start++;
      else
	f->file_selector.dir_pos++;
      draw_dir(f);
      break;
    case XK_Up :
      if (f->file_selector.dir_start+f->file_selector.dir_pos==0)
	break;
      if (f->file_selector.dir_pos==0)
	f->file_selector.dir_start--;
      else
	f->file_selector.dir_pos--;
      draw_dir(f);
      break;
    case XK_Page_Down :
      if (f->file_selector.dir_start+f->file_selector.dir_pos==f->file_selector.nb_dirs-1)
	break;
      if (f->file_selector.dir_pos<15)
	f->file_selector.dir_pos=15;
      else
	f->file_selector.dir_start+=15;
      if (f->file_selector.dir_start+f->file_selector.dir_pos>=f->file_selector.nb_dirs)
	f->file_selector.dir_pos=f->file_selector.nb_dirs-f->file_selector.dir_start-1;
      draw_dir(f);
      break;
    case XK_Page_Up :
      if (f->file_selector.dir_start+f->file_selector.dir_pos==0)
	break;
      if (f->file_selector.dir_pos>0)
	f->file_selector.dir_pos=0;
      else {
	f->file_selector.dir_start-=15;
	if (f->file_selector.dir_start<0)
	  f->file_selector.dir_start=0;
      }
      draw_dir(f);
      break;
    case XK_Return :
      /* this is barbarian code, don't read it, please, am shamefull */
      /* first, see if we can do a chdir */
      if (chdir(f->file_selector.dirs[f->file_selector.dir_pos+f->file_selector.dir_start])==-1) {
	XBell(f->X_ecran, 0);
	break;
      }
      /* done, let's kill all dirs, all files, and redraw all */
      reset_file_dir(f);
      get_directories_files(&f->file_selector.files, &f->file_selector.dirs,
			    &f->file_selector.nb_files, &f->file_selector.nb_dirs);
      f->file_selector.dir_pos=f->file_selector.dir_start=
      f->file_selector.file_pos=f->file_selector.file_start=0;
      draw_dir_and_file_names(f);
      /* now, put the name of current dir in out_name and redraw it */
      if (f->file_selector.out_name)
	free(f->file_selector.out_name);
      f->file_selector.out_name=getcwd((char *)0, 0);
      if (f->file_selector.out_name)
	draw_out_name(f);
      break;
  }
}

void interface_win_sel_2(INTERFACE *f, KeySym ks)
{
  switch(ks) {
    case XK_Down :
      /* this test is enough, it can't be higher (0>-1), because if nb_files==0, we never will be here */
      if (f->file_selector.file_start+f->file_selector.file_pos==f->file_selector.nb_files-1)
	break;
      if (f->file_selector.file_pos==15)
	f->file_selector.file_start++;
      else
	f->file_selector.file_pos++;
      draw_file(f);
      break;
    case XK_Up :
      if (f->file_selector.file_start+f->file_selector.file_pos==0)
	break;
      if (f->file_selector.file_pos==0)
	f->file_selector.file_start--;
      else
	f->file_selector.file_pos--;
      draw_file(f);
      break;
    case XK_Page_Down :
      if (f->file_selector.file_start+f->file_selector.file_pos==f->file_selector.nb_files-1)
	break;
      if (f->file_selector.file_pos<15)
	f->file_selector.file_pos=15;
      else
	f->file_selector.file_start+=15;
      if (f->file_selector.file_start+f->file_selector.file_pos>=f->file_selector.nb_files)
	f->file_selector.file_pos=f->file_selector.nb_files-f->file_selector.file_start-1;
      draw_file(f);
      break;
    case XK_Page_Up :
      if (f->file_selector.file_start+f->file_selector.file_pos==0)
	break;
      if (f->file_selector.file_pos>0)
	f->file_selector.file_pos=0;
      else {
	f->file_selector.file_start-=15;
	if (f->file_selector.file_start<0)
	  f->file_selector.file_start=0;
      }
      draw_file(f);
      break;
    case XK_Return : {
      char *tmp;
      if (f->file_selector.out_name)
	free(f->file_selector.out_name);
      f->file_selector.out_name=getcwd((char *)0, 0);
      if (!f->file_selector.out_name) {
	fprintf(stderr, "file_selector.interface_win_sel_2 error with getcwd\n");
	draw_out_name(f);
	break;
      }
      tmp=realloc(f->file_selector.out_name, (strlen(f->file_selector.out_name)+
		  strlen(f->file_selector.files[f->file_selector.file_start+f->file_selector.file_pos])+2)*sizeof(char));
      if (!tmp) {
	fprintf(stderr, "file_selector.interface_win_sel_2 error with realloc\n");
	draw_out_name(f);
	break;
      }
      f->file_selector.out_name=tmp;
      strcpy(&f->file_selector.out_name[strlen(f->file_selector.out_name)+1],
	     f->file_selector.files[f->file_selector.file_start+f->file_selector.file_pos]);
      f->file_selector.out_name[strlen(f->file_selector.out_name)]='/';
      draw_out_name(f);
      return;
    }
  }
}

void (*enter_win_sel[5])(INTERFACE *)={
  enter_win_sel_0,
  enter_win_sel_1,
  enter_win_sel_2,
  enter_win_sel_3,
  enter_win_sel_4,
};

void (*exit_win_sel[5])(INTERFACE *)={
  exit_win_sel_0,
  exit_win_sel_1,
  exit_win_sel_2,
  (void (*)(INTERFACE *))0,
  (void (*)(INTERFACE *))0
};

void (*interface_win_sel[5])(INTERFACE *, KeySym)={
  interface_win_sel_0,
  interface_win_sel_1,
  interface_win_sel_2,
  (void (*)(INTERFACE *, KeySym))0,
  (void (*)(INTERFACE *, KeySym))0
};

int init_file_selector_graphic(INTERFACE *f)
{
  /* once again, no test are done to verify everything was allright */
  f->file_selector.window=XCreateSimpleWindow(f->X_ecran, DefaultRootWindow(f->X_ecran),
					      0, 0, 500, HAUTEUR, 3, 3,
					      BlackPixel(f->X_ecran, DefaultScreen(f->X_ecran)));
  XMapWindow(f->X_ecran, f->file_selector.window);
  XSelectInput(f->X_ecran, f->file_selector.window, KeyPressMask|ExposureMask);
  f->file_selector.pixmap=XCreatePixmap(f->X_ecran, DefaultRootWindow(f->X_ecran), 500, HAUTEUR,
					DefaultDepth(f->X_ecran, DefaultScreen(f->X_ecran)));
  draw_box(f->X_ecran, f->file_selector.pixmap, f->context, 0, 0, 500, HAUTEUR, BLACK);

  draw_3d_box(f->X_ecran, f->file_selector.pixmap, f->context, 4, 4, 500-8, 30, GREY, UP_3D);
  draw_3d_box(f->X_ecran, f->file_selector.pixmap, f->context, 4, 38, (500-8)/2-2, 14*16+4+20, GREY, UP_3D);
  draw_3d_box(f->X_ecran, f->file_selector.pixmap, f->context, 4+(500-8)/2+2, 38, (500-8)/2-2, 14*16+4+20, GREY, UP_3D);
  //  draw_3d_box(f->X_ecran, f->file_selector.pixmap, f->context, 4, 4, 500-8, 30, GREY, UP_3D);
  //draw_3d_box(f->X_ecran, f->file_selector.pixmap, f->context, 4, 4, 500-8, 30, GREY, UP_3D);
  draw_string(f->X_ecran, f->file_selector.pixmap, f->context, 6, 14,
	      "Current name :", f->little, BLACK);
  draw_box(f->X_ecran, f->file_selector.pixmap, f->context, 10, 28-10-1, 500-8-6-4, 13, LIGHTBLUE);
  draw_string(f->X_ecran, f->file_selector.pixmap, f->context, 6+5, 38+10+5,
	      "Sub-directories :", f->little, BLACK);
  draw_string(f->X_ecran, f->file_selector.pixmap, f->context, 6+(500-8)/2+2+5, 38+10+5,
	      "Files :", f->little, BLACK);
  draw_3d_box(f->X_ecran, f->file_selector.pixmap, f->context, 250-70, 38+14*16+4+20+4, 60, 20, LIGHTBLUE, UP_3D);
  draw_3d_box(f->X_ecran, f->file_selector.pixmap, f->context, 250+10, 38+14*16+4+20+4, 60, 20, LIGHTBLUE, UP_3D);
  draw_string(f->X_ecran, f->file_selector.pixmap, f->context, 250-60+6+4,  38+14*16+4+20+4+14, "OK", f->medium, BLACK);
  draw_string(f->X_ecran, f->file_selector.pixmap, f->context, 250+10+3,  38+14*16+4+20+4+14, "Cancel", f->medium, BLACK);
  {
    /* put the name of the window (i am not sure of this code, shoud read books to know if it's correct */
    XTextProperty proper;
    char *z="Xsoundtrack : file selector";
    if (!XStringListToTextProperty(&z, 1, &proper))
      fprintf(stderr, "file_selector.init_file_selector_graphic error with XStringToTextProperty\n");
    else {
      XSetWMName(f->X_ecran, f->file_selector.window, &proper);
      XFree(proper.value);
    }
  }
  around_file_selector(f, 0, IN);
  return 1;
}

int init_file_selector_data(INTERFACE *f)
{
  if (f->file_selector.out_name) {
    free(f->file_selector.out_name);
    f->file_selector.out_name=(char *)0;
  }
  f->file_selector.in_use=1;
  f->file_selector.cur_pos=0;
  f->file_selector.down_0=1;
  f->file_selector.mode=MOVING;
  f->file_selector.files=(char **)0;
  f->file_selector.dirs=(char **)0;
  f->file_selector.nb_files=0;
  f->file_selector.nb_dirs=0;
  f->file_selector.file_start=0;
  f->file_selector.dir_start=0;
  f->file_selector.file_pos=0;
  f->file_selector.dir_pos=0;
  get_directories_files(&f->file_selector.files, &f->file_selector.dirs,
			&f->file_selector.nb_files, &f->file_selector.nb_dirs);
  _draw_dir_and_file_names(f);
  _draw_out_name(f);

  return 1;
}

void close_file_selector(INTERFACE *f)
{
  XDestroyWindow(f->X_ecran, f->file_selector.window);
  XFreePixmap(f->X_ecran, f->file_selector.pixmap);
  reset_file_dir(f);
}

/* draw the window, return 0 if pb 1 otherwise */
int init_file_selector(INTERFACE *f, char *current_dir)
{
  /* defined in main.c */
  void real_main(INTERFACE *);

  /* move to current_dir */
  if (chdir(current_dir)==-1) {
    perror("file_selector.init_file_selector error\nperror says");
    return 0;
  }

  if (!init_file_selector_graphic(f))
    return 0;

  if (!init_file_selector_data(f)) {
    close_file_selector(f);
    return 0;
  }

  while(f->file_selector.in_use)
    real_main(f);

  close_file_selector(f);

  return 1;
}

void draw_file_selector(INTERFACE *f)
{
  XCopyArea(f->X_ecran, f->file_selector.pixmap, f->file_selector.window, f->context, 0, 0, 500, HAUTEUR, 0, 0);
}

void keypress_file_selector(INTERFACE *f, XKeyEvent *ev)
{
  char dum[50];
  KeySym ks;

  XLookupString(ev, dum, 50, &ks, (XComposeStatus *)0);

  if (ks==NoSymbol) {
    fprintf(stderr, "file_selector.keypress_file_selector : warning ks=NoSymbol\n");
    return;
  }

  /* escape has a global signification, so a global handling */
  if(ks==XK_Escape) {
    if (f->file_selector.mode==EDIT) {
      f->file_selector.mode=MOVING;
      exit_win_sel[f->file_selector.cur_pos](f);
      return;
    }
    return;
  }

  if (ks==XK_Tab) {
    if (f->file_selector.cur_pos==1) {
      around_file_selector(f, 1, OUT);
      around_file_selector(f, 2, IN);
      f->file_selector.cur_pos=2;
      if (f->file_selector.mode==EDIT) {
	exit_win_sel_1(f);
	enter_win_sel_2(f);
      }
      return;
    }
    if (f->file_selector.cur_pos==2) {
      around_file_selector(f, 2, OUT);
      around_file_selector(f, 1, IN);
      f->file_selector.cur_pos=1;
      if (f->file_selector.mode==EDIT) {
	exit_win_sel_2(f);
	enter_win_sel_1(f);
      }
      return;
    }
  }

  if (f->file_selector.mode==EDIT)
    return interface_win_sel[f->file_selector.cur_pos](f, ks);

  switch(ks) {
    case XK_space :
    case XK_Return :
      if (f->file_selector.cur_pos==1 && !f->file_selector.nb_dirs) {
	XBell(f->X_ecran, 0);
	break;
      }
      if (f->file_selector.cur_pos==2 && !f->file_selector.nb_files) {
	XBell(f->X_ecran, 0);
	break;
      }
      f->file_selector.mode=EDIT;
      enter_win_sel[f->file_selector.cur_pos](f);
      break;
    case XK_Left :
      switch(f->file_selector.cur_pos) {
	default :
	  break;
	case 2 :
	  around_file_selector(f, 2, OUT);
	  around_file_selector(f, 1, IN);
	  f->file_selector.cur_pos=1;
	  break;
	case 4 :
	  around_file_selector(f, 4, OUT);
	  around_file_selector(f, 3, IN);
	  f->file_selector.cur_pos=3;
	  break;
      }
      break;
    case XK_Right :
      switch(f->file_selector.cur_pos) {
	default :
	  break;
	case 1 :
	  around_file_selector(f, 1, OUT);
	  around_file_selector(f, 2, IN);
	  f->file_selector.cur_pos=2;
	  break;
	case 3 :
	  around_file_selector(f, 3, OUT);
	  around_file_selector(f, 4, IN);
	  f->file_selector.cur_pos=4;
	  break;
      }
      break;
    case XK_Up :
      switch(f->file_selector.cur_pos) {
	case 0 :
	  break;
	case 1 :
	  around_file_selector(f, 1, OUT);
	  around_file_selector(f, 0, IN);
	  f->file_selector.cur_pos=0;
	  f->file_selector.down_0=1;
	  break;
	case 2 :
	  around_file_selector(f, 2, OUT);
	  around_file_selector(f, 0, IN);
	  f->file_selector.cur_pos=0;
	  f->file_selector.down_0=2;
	  break;
	case 3 :
	  around_file_selector(f, 3, OUT);
	  around_file_selector(f, 1, IN);
	  f->file_selector.cur_pos=1;
	  break;
	case 4 :
	  around_file_selector(f, 4, OUT);
	  around_file_selector(f, 2, IN);
	  f->file_selector.cur_pos=2;
	  break;
      }
      break;
    case XK_Down :
      switch(f->file_selector.cur_pos) {
	default :
	  break;
	case 0 :
	  around_file_selector(f, 0, OUT);
	  around_file_selector(f, f->file_selector.down_0, IN);
	  f->file_selector.cur_pos=f->file_selector.down_0;
	  break;
	case 1 :
	  around_file_selector(f, 1, OUT);
	  around_file_selector(f, 3, IN);
	  f->file_selector.cur_pos=3;
	  break;
	case 2 :
	  around_file_selector(f, 2, OUT);
	  around_file_selector(f, 4, IN);
	  f->file_selector.cur_pos=4;
	  break;
      }
      break;
  }
}

void file_selector_handle_event(INTERFACE *f, XEvent *ev)
{
  switch(ev->type) {
    case Expose :
      draw_file_selector(f);
      break;
    case KeyPress :
      keypress_file_selector(f, &ev->xkey);
      break;
  }
}
