/*
 * This file is part of xsoundtrack - a sound tracker for Linux.
 * Sed - july august 1998.
 * This is total free software.
 *-------------------------------------
 * The file contains a mixer, ie routines necessary to write the sound to the soundcard.
 */

#include "mixer.h"

#include "xm.h"
#include "card.h"
#include "live_player.h"

/* maybe not all those include are necessary, due to coding-process, which cause destruction of code after its need */
#include <stdio.h>
#include <string.h>
#include <math.h>

/* the channels */
/* have moved in each xm */

/* buffer for mixing in 32b */
int mixer_soft_buffer[1024*2/4];
int mixer_pos;
/* buffer for output in the soundcard */
/* one day, will have to let user choose between != resolution (8, 16b, 44100Hz or something else, mono/stereo) */
char mixer_buffer_soundcard[1024];

int init_mixer(int rate, int precision, int mono_stereo)
{
  /* well, for the first player, we gonna fix the things, so only 44100/16/stereo (incremental coding) */
  /* memset(channels, 0, sizeof(channels)); */
  return 1;
}

/* this function inits some global vars in the xm-module */
void init_xm_data(XM *module)
{
/*  int new_line(XM *); */
  module->current_pattern=0;
  module->current_line=0; /*-1;*/     /* -1 because the mixer will realise it has no more byte to handle, so increment
				* the current line, which shall reach 0
				*/
  module->current_tempo=module->default_tempo; /* why did they call it tempo ? */
  /*  module->speed=32760;*/ /* one big value to directly change the line in new_line, bad method, will have to handle
			* it differently
			*/
  /*  fprintf(stderr, "mixer.init_xm_data don't forget to handle the very first line differently or what...\n");*/

  module->bpm=module->default_bpm;
/*  module->to_handle=new_line(module); */ /* we gonna do that only when necessary */
  module->nb_bytes=(44100L*8*5/(module->bpm*2))&(~7); /* the & ~7 because a datum is 8 bytes long, so we want a 8 bytes
						       * alignment
						       */
  module->global_volume=64;

  module->current_line--;
  module->speed=32768;
  module->to_handle=new_line(module);
  module->new_line=module->new_pattern=0;
}

/* you give an instrument, a note, and it tells you what increment to take.
 * The first 32b are integere-increment, the next 32b are the after the dot
 * example : if increment=123.456, then the 32b first are 123 and the 32next are 2^32*.456
 */
long long get_increment(instrument *inst, int note)
{
  unsigned long long ret;
  long double period, freq, fl;
  sample *the_sample=&inst->sample_headers[(int)inst->sample_number[note]];
  char fine=the_sample->sample_fine_tune;//-128;
  note+=the_sample->sample_relative_note_number;
  period=10.*12*16*4- (long double)note*16*4 - (long double)fine/2.;
  freq=8363. * pow(2, (6*12*16*4 - period)/(12*16*4));
  //fprintf(stderr, "note = %d freq = 8363*2^(%Lf) = %Lf\t", note, (6*12*16*4 - period)/(12*16*4), freq);
  /* if 16b sample, then 2 bytes per datum, so a double freq */
  if (the_sample->sample_type & 0x10)
    freq*=2.;
  freq/=44100.;
  //fprintf(stderr, "freq = %Lf\t", freq);
  fl=floor(freq);
  //fprintf(stderr, "fl = %Lf\t", fl);
  freq-=fl;
  //fprintf(stderr, "freq = %Lf\t", freq);
  freq*=(unsigned long long)0xffffffff;
  freq=rint(freq);
  ret=(((unsigned long long)fl)<<32)|((unsigned long)freq);
  //fprintf(stderr, "ret normal = %lld ret haut = %lld ret bas = %lld\n", ret, ret>>32, ret & 0xffffffff);
  return ret;
}

/* this function handle the effects (ie column 3 & 4 of a note in the sample editor) */
void handle_effects(XM *xm)
{
  int i;
  channel *channels=xm->channels;
  int tick=xm->speed;

  for (i=0; i<xm->nb_channels; i++) {
    int newvol=channels[i].volume, newpan=channels[i].panning;
    int v=(channels[i].volume_effect)&255;
    int e=(channels[i].effect_type)&255;
    int p=(channels[i].effect_parameter)&255;

#if 0
/* we may have unused channels that have effects (main case: effect f, set speed/bpm), so fuck this test */
    if (!channels[i].in_use)
      continue;
#endif

    /* the spec. says to handle the volume column effect first */
    /* BUT it says that it is overwriten by effect column effect, so
     * we've got to use local variables */
    switch ((v>>4)&0xf) {
      case 0 :
	/* is it necessary to deal with this value ? Will the compiler
	 * optimize stuff with or without it ??? (yes, i am a very bad coder, don't even know that)
	 */
	break;
      case 1 :
      case 2 :
      case 3 :
      case 4 :
      case 5 :
	/* set volume */
	if (v>0x50)
	  break;
	if (tick)
	  break;
	newvol=v-0x10;
	break;
      case 6 :
	/* volume slide down */
	if (!tick)
	  break;
	newvol=channels[i].volume-(v&15);
	if (newvol<0)
	  newvol=0;
	break;
      case 7 :
	/* volume slide up */
	if (!tick)
	  break;
	newvol=channels[i].volume+(v&15);
	if (newvol>0x40)
	  newvol=0x40;
	break;
      case 8 :
	/* fine volume slide down */
	if (tick!=1)
	  break;
	newvol=channels[i].volume-(v&15);
	if (newvol<0)
	  newvol=0;
	break;
      case 9 :
	/* fine volume slide up */
	if (tick!=1)
	  break;
	newvol=channels[i].volume+(v&15);
	if (newvol>0x40)
	  newvol=0x40;
	break;
      case 0xa :
	/* set vibrato speed */
	/* TODO */
	break;
      case 0xb :
	/* vibrato */
	/* TODO */
	break;
      case 0xc :
	/* set panning */
	if (tick)
	  break;
	newpan=v<<4;
	break;
      case 0xd :
	/* panning slide left */
	if (tick)
	  break;
	newpan=channels[i].panning-(v&15);
	if (newpan<0)
	  newpan=0;
	break;
      case 0xe :
	/* panning slide right */
	if (tick)
	  break;
	newpan=channels[i].panning+(v&15);
	if (newpan>255)
	  newpan=255;
	break;
      case 0xf :
	/* tone porta */
	/* TODO */
	break;
    }

    /* then, the effect column effect (nice english i have, huh ?) */
    switch (e) {
      case -1 :
	/* once again, what is best ? put this case or not ? */
	break;
      case 8 :
	/* set panning position */
	if (tick)
	  break;
	newpan=p;
	break;
      case 10 : {
	/* volume slide (up or down) */
	int slide;
	if (!tick)
	  break;
	if (p)
	  channels[i].old_volume_slide=p;
	slide=channels[i].old_volume_slide;
	/* slide up has priority, dno't know if it's true. By the way, with Xsoundtrack,
	 * this case is impossible
	 */
	if ((slide&0xf0)&&(slide&0x0f))
	  slide&=0xf0;
	if (slide&0xf0) {
	  /* slide up */
	  slide>>=4;
	  newvol=channels[i].volume+slide;
	  if (newvol>0x40)
	    newvol=0x40;
	} else {
	  /* slide down */
	  newvol=channels[i].volume-slide;
	  if (newvol<0)
	    newvol=0;
	}
	break;
      }
      case 11 :
	/* position jump */
	/* well, in fact, it's got to jump at the end of the row, so it's handled
	 * in new_line, we just inform it it has to jump and where to go
	 */
	if (tick)
	  break;
	xm->next_line_is_a_jump++;
	xm->jump_to=p;
	break;
      case 12 :
	/* set volume */
	if (tick)
	  break;
	newvol=p;
	break;
      case 13 :
	/* jump to next pattern at specified row (can mix with position jump, sanx to Ripper for walker.xm) */
	if (tick)
	  break;
	xm->next_line_is_a_set_row++;
	xm->next_row=p;
	break;
      case 14 :
	/* multiple stuff */
	switch((p&0xf0)>>4) {
	  case 6 : {
	    int param=p&15;
	    /* set loop point / loop to loop point */
	    if (tick)
	      break;
	    /* if a e60 follows another e60 in the same channel, ignore it */
	    /* nope, take the last one, better */
	    if (!param) {
	      if (!channels[i].has_looped) {
		if (!channels[i].loop_point_is_set) {
		  channels[i].loop_point_is_set++;
		  channels[i].looped_nb=0;
		  channels[i].loop_row=xm->current_line;
		  channels[i].loop_pattern=xm->current_pattern;
		}
	      } else
		channels[i].has_looped=0;
	    } else {
	      /* if no e60 set before or bad e60 set (wrong pattern), do nothing */
	      if (channels[i].loop_point_is_set && channels[i].loop_pattern==xm->current_pattern) {
		channels[i].loop_max=param;
		xm->next_line_is_a_loop++;
	      }
	    }
	  }
	  case 0xA : {
	    /* fine volume slide up */
	    int slide;
	    int param=p&15;
	    if (tick!=1)
	      break;
	    if (param)
	      channels[i].old_fine_volume_slide_up=param;
	    slide=channels[i].old_fine_volume_slide_up;
	    newvol=channels[i].volume+slide;
	    if (newvol>0x40)
	      newvol=0x40;
	    break;
	  }
	  case 0xB : {
	    /* fine volume slide down */
	    int slide;
	    int param=p&15;
	    if (tick!=1)
	      break;
	    if (param)
	      channels[i].old_fine_volume_slide_down=param;
	    slide=channels[i].old_fine_volume_slide_down;
	    newvol=channels[i].volume-slide;
	    if (newvol<0)
	      newvol=0;
	    break;
	  }
	  case 0xC :
	    /* note cut */
	    if (tick==(p&15))
	      newvol=0;
	    break;
	}
	break;
      case 15 :
	/* set speed/bpm */
	if (tick)
	  break;
	if (p==0) /* i think ft2 uses this value too, but spec. say no */
	  break;
	if (p<=0x1f) {
	  /* set speed */
	  /* we have 2 vars, bad, should remove one of them */
	  xm->current_tempo=xm->default_tempo=p;
	} else {
	  /* set bpm */
	  /* same as above for the vars */
	  xm->bpm=xm->default_bpm=p;
	}
	/* we must redraw things */
	xm->new_tempo++;
	break;
      case 16 :
	/* set global volume */
	if (tick)
	  break;
	xm->global_volume=p>=0x40?0x40:p;
	break;
      case 17 : {
	/* global volume slide */
	int slide;
	if (!tick)
	  break;
	if (p)
	  channels[i].old_global_volume_slide=p;
	slide=channels[i].old_global_volume_slide;
	if ((slide&0xf0)&&(slide&0x0f))
	  slide&=0xf0;
	if (slide&0xf0) {
	  /* slide up */
	  slide>>=4;
	  xm->global_volume+=slide;
	  if (xm->global_volume>0x40)
	    xm->global_volume=0x40;
	} else {
	  /* slide down */
	  xm->global_volume-=slide;
	  if (xm->global_volume<0)
	    xm->global_volume=0;
	}
	break;
      }
      case 25 : {
	/* panning slide */
	int slide;
	if (!tick)
	  break;
	if (p)
	  channels[i].old_panning_slide=p;
	slide=channels[i].old_panning_slide;
	if ((slide&0xf0)&&(slide&0x0f))
	  slide&=0xf0;
	if (slide&0xf0) {
	  /* slide up */
	  slide>>=4;
	  newpan=channels[i].volume+slide;
	  if (newpan>0x40)
	    newpan=0x40;
	} else {
	  /* slide down */
	  newpan=channels[i].volume-slide;
	  if (newpan<0)
	    newpan=0;
	}
	break;
      }
    }

    channels[i].volume=newvol;
    channels[i].panning=newpan;
  }
}

int find_loop_channel(XM *module)
{
  int i;
  int good_channel=-1;
  int the_pattern=module->patterns_order[module->current_pattern];
  int nb_rows=module->patterns_nb_rows[the_pattern];
  note *cur_note=module->patterns[the_pattern]+module->current_line;
  /* bad things could happen if both a loop and a jump were performed at the same row.
   * Should the loop have priority ? Will have to ask to Ripper.
   * for now, we suppose it is not possible to have both of them
   */
  /* find the right channel to loop (the one that jumps the less) */
  for (i=0; i<module->nb_channels; i++, cur_note+=nb_rows) {
    /* must loop, and must have the e6x in it's effect column */
    if (!module->channels[i].loop_point_is_set)
      continue;
    if (cur_note->effect_type!=0xe || (cur_note->effect_parameter&0xf0)!=6<<4 || !(cur_note->effect_parameter&15))
      continue;
    /* sanity check (same pattern, loop point upper) */
    if (module->current_pattern!=module->channels[i].loop_pattern || module->current_line<=module->channels[i].loop_row)
      continue;
    if (good_channel==-1) {
      good_channel=i;
      continue;
    }
    /* compare the loop points, take the biggest (if ==, well, gonna take the first) */
    if (module->channels[i].loop_row>module->channels[good_channel].loop_row)
      good_channel=i;
  }
  return good_channel;
}

/* this function will read a new line in the sample, change global-vars consequently */
/* return the number of bytes the tick is (so, it will be called not only when we change line
 * but when we change tick, it's possible for a line to be more than one tick (up to 32 i think),
 * so it's badly named.
 * Nota : if we are dumping, we won't loop, so we return 0 at the end.
 */
int new_line(XM *module)
{
  channel *channels=module->channels;
  unsigned int the_pattern;
  int i, nb_rows;
  note *cur_note;

  /* here, we maybe only have to go on for one tick */

  if (module->speed<module->current_tempo)
    /* yes, it only was one tick, so don't change the line, but handle effects, so a little goto (i love those gotos) */
    goto dont_change_line;

  /* here, we gotta change of line */
  module->speed=0;

  /* well, don't know how to handle it... by now, i only need to fill the 32 channels with :
   *   - is there a note or no ?
   *   - if yes, then fill the data in the channel :
   *          - current pointer of the sample
   *          - increment (64b, long long, hehe) (hehe, i realized it's not so good, damned)
   *         # - that's all folks, no loop, no panning, no volume. Just to see if it works so
   */

  /* ok, don't jump anywhere if in edit mode when not playing (strange behavior, but don't
   * exactly know what semantic i want, so the easiest hack i found is that)
   */
  /* shitty if else if here */
  if ((!module->next_line_is_a_jump && !module->next_line_is_a_set_row && !module->next_line_is_a_loop) ||
      !(module->mix&MODULE_PLAY)) {
    module->next_line_is_a_jump=0;
    module->next_line_is_a_set_row=0;
    module->next_line_is_a_loop=0;
    module->current_line++;
  } else {
    if (module->next_line_is_a_loop) {
      int good_channel;
      /* deal the loop */
      /* we search a loop channel. If we find one, we may be finish with it, so close it
       * and keep searching.
       */
      do {
	good_channel=find_loop_channel(module);
	if (good_channel!=-1) {
	  if (++(module->channels[good_channel].looped_nb)<=module->channels[good_channel].loop_max) {
	    module->current_line=module->channels[good_channel].loop_row;
	    module->channels[good_channel].has_looped++;
	    break;
	  }
	  /* we looped all we had to for this channel, clean the loop for this channel*/
	  module->channels[good_channel].loop_point_is_set=0;
	}
      } while (good_channel!=-1);
      if (good_channel==-1)
	module->current_line++;
      module->next_line_is_a_loop=0;
    } else {
      int error=0;
      /* if we jump over the song, silently ignore this request */
      if (module->next_line_is_a_jump) {
	if (module->jump_to<module->song_length) {
	  module->current_pattern=module->jump_to;
	  module->current_line=0;
	  module->new_pattern=1;
	} else {
	  error++;
	  module->next_line_is_a_jump=0;
	}
      }
      if (module->next_line_is_a_set_row) {
	/* two cases : we have a jump or not */
	if (module->next_line_is_a_jump) {
	  /* if next_row out of pattern, ignore it (stay to 0) */
	  if (module->patterns_nb_rows[(int)module->patterns_order[module->current_pattern]]>module->next_row)
	    module->current_line=module->next_row;
	} else {
	  /* if next pattern doesn't exist, or next_row out of pattern, ignore it */
	  if (module->current_pattern+1<=module->song_length-1 &&
	      module->patterns_nb_rows[(int)module->patterns_order[module->current_pattern]]>module->next_row) {
	    module->current_pattern++;
	    module->current_line=module->next_row;
	    module->new_pattern=1;
	    error=0;
	  } else
	    error++;
	}
      }
      module->next_line_is_a_jump=0;
      module->next_line_is_a_set_row=0;
      if (error)
	module->current_line++;
    }
  }

  module->new_line=1;

  if (module->current_line >= module->patterns_nb_rows[(int)module->patterns_order[module->current_pattern]]) {
    module->current_pattern++;
    module->new_pattern=1;
    module->current_line=0;
    if (module->current_pattern>=module->song_length) {
      /* it's over, we loop, except if dumping */
      if (module->do_we_dump)
	return 0;
      module->current_pattern=module->restart;
    }
  }

  /***************** here, i have to not forget to test things, it core-dumps ******************/
  /* done, it does not core anymore (but am i really sure of it ?) */
  the_pattern=module->patterns_order[module->current_pattern];
  nb_rows=module->patterns_nb_rows[the_pattern];
  if (!module->patterns) {
    fprintf(stderr, "error, pattern not existing\n");
    /* this exit(1) is not pretty, should take it away */
    exit(1);
  }
  cur_note=module->patterns[the_pattern]+module->current_line;
  /* ok, let's run through the channels */
  /* we just gonna put the instrument to play, set current position to 0, that's all */
  for (i=0; i<module->nb_channels; i++, cur_note+=nb_rows) {
    int instru;
    channels[i].volume_effect=cur_note->volume;
    channels[i].effect_type=cur_note->effect_type;
    channels[i].effect_parameter=cur_note->effect_parameter;
    /* if channel is closed, pass it over */
    if (channels[i].closed)
      continue;
    /* test cut-note, if envelop is there with sustain or without it !! */
    if (cur_note->note==96) {
      if (channels[i].instrument && (channels[i].instrument->volume_type & 1))
	channels[i].cut_key=1;
      else
	channels[i].in_use=0;
      continue;
    }
    if (cur_note->note==0xff)
      continue; /*goto ici_ici_ici;*/
    if (cur_note->instrument!=-1)
      channels[i].last_instrument=cur_note->instrument;
    instru=channels[i].last_instrument;
    /* we load a new note, set all parameters */
    if (!(channels[i].instrument=module->instruments[instru])) {
      channels[i].in_use=0;
      continue;
    }
    channels[i].sample=&channels[i].
      instrument->sample_headers[(int)channels[i].instrument->sample_number[(int)cur_note->note]];
    if (channels[i].sample->sample_data==(char *)0) {
      channels[i].in_use=0;
      continue;
    }
    channels[i].sample_position=0;
    channels[i].increment=get_increment(module->instruments[instru], cur_note->note);
    channels[i].in_use=1;
    channels[i].volume_point=0;
    channels[i].panning_point=0;
    channels[i].envelop_volume=channels[i].instrument->volume_type & 1 ? channels[i].instrument->volume_envelop[0][1] : 64;
    channels[i].envelop_panning=channels[i].instrument->panning_type & 1 ?
      channels[i].instrument->panning_envelop[0][1] : 32;
    channels[i].volume=channels[i].sample->sample_volume;
    channels[i].panning=channels[i].sample->sample_panning;
    channels[i].cut_key=0;
#if 0
/* volume setting is sort of an effect, handled later, with other effects */
ici_ici_ici:
    if (cur_note->volume>=0x10 && cur_note->volume<=0x50)
      channels[i].volume=cur_note->volume-0x10;
#endif
  }
  //fprintf(stderr, "\n");
  //new_line();
  //exit(1);

  /* we've changed the line, let's handle the effects */

dont_change_line:

  /* should it be here or after the envelop dealing ?? */
  /* gonna put it before because of the volume setting, and the global calculus
   * has to be done after */
  handle_effects(module);

  /* let's handle the volume / panning */
  for (i=0; i<module->nb_channels; i++) {
    instrument *ins;

    if (!channels[i].in_use)
      continue;

    /* first, set the left/right volume according to the current volume / panning setting */
    ins=channels[i].instrument;

    {
      /* calculate final panning according to xm.txt */
      int pan=(unsigned char)channels[i].panning;
      int envpan=channels[i].envelop_panning;
      int final_panning=pan + ((envpan-32)*(128-abs(pan-128)))/32;

      channels[i].left_volume=(channels[i].volume*module->global_volume*
			       channels[i].envelop_volume*(255-final_panning))>>(6+6+8);
      channels[i].right_volume=(channels[i].volume*module->global_volume*
				channels[i].envelop_volume*final_panning)>>(6+6+8);
    }

    /* now, move the volume point */
    if (!(ins->volume_type & 1))
      goto deal_with_panning;

    /* are we on the sustain point ? */
    if ((!channels[i].cut_key) && (ins->volume_type & 2) &&
	(channels[i].volume_point == ins->volume_envelop[(int)ins->volume_sustain_point][0]))
      goto deal_with_panning;

    /* was it the loop point ? */
    if (ins->volume_type & 4) {
      if (channels[i].volume_point == ins->volume_envelop[(int)ins->volume_loop_end][0]) {
	channels[i].volume_point=ins->volume_envelop[(int)ins->volume_loop_start][0];
	channels[i].envelop_volume=ins->volume_envelop[(int)ins->volume_loop_start][1];
	goto deal_with_panning;
      }
    }

    /* or the end ? */
    if (channels[i].volume_point == ins->volume_envelop[ins->nb_volume_points-1][0])
      goto deal_with_panning;

    channels[i].volume_point++;
    /* set the envelop volume (ugly code) */
    {
      int right;
      int x, x1, y1, x2, y2;
      for (right=1; right<ins->nb_volume_points; right++)
	if (ins->volume_envelop[right][0]>=channels[i].volume_point)
	  break;
      /* right is the first point on the right of the current volume point, we interpolate to find the new volume */
      /*
	          /(x2, y2)
		 /
		/ (x, Y)
	       /(x1, y1)
           (x-x1)/(x2-x1)=(Y-y1)/(y2-y1)
	   Y=y1 + (x-x1)/(x2-x1)*(y2-y1)
       */
      x=channels[i].volume_point;
      x1=ins->volume_envelop[right-1][0];
      y1=ins->volume_envelop[right-1][1];
      x2=ins->volume_envelop[right][0];
      y2=ins->volume_envelop[right][1];
      x-=x1;
      x2-=x1;
      y2-=y1;
      channels[i].envelop_volume = y1 + x*y2/x2;
    }

deal_with_panning:
    /* now, move the panning point */
    if (!(ins->panning_type & 1))
      continue;

    /* are we on the sustain point ? */
    if ((!channels[i].cut_key) && (ins->panning_type & 2) &&
	(channels[i].panning_point == ins->panning_envelop[(int)ins->panning_sustain_point][0]))
      continue;

    /* was it the loop point ? */
    if (ins->panning_type & 4) {
      if (channels[i].panning_point == ins->panning_envelop[(int)ins->panning_loop_end][0]) {
	channels[i].panning_point=ins->panning_envelop[(int)ins->panning_loop_start][0];
	channels[i].envelop_panning=ins->panning_envelop[(int)ins->panning_loop_start][1];
	continue;
      }
    }

    /* or the end ? */
    if (channels[i].panning_point == ins->panning_envelop[ins->nb_panning_points-1][0])
      continue;

    channels[i].panning_point++;
    /* set the envelop panning (ugly code) */
    {
      int right;
      int x, x1, y1, x2, y2;
      for (right=1; right<ins->nb_panning_points; right++)
	if (ins->panning_envelop[right][0]>=channels[i].panning_point)
	  break;
      /* right is the first point on the right of the current panning point, we interpolate to find the new panning */
      /*
	          /(x2, y2)
		 /
		/ (x, Y)
	       /(x1, y1)
           (x-x1)/(x2-x1)=(Y-y1)/(y2-y1)
	   Y=y1 + (x-x1)/(x2-x1)*(y2-y1)
       */
      x=channels[i].panning_point;
      x1=ins->panning_envelop[right-1][0];
      y1=ins->panning_envelop[right-1][1];
      x2=ins->panning_envelop[right][0];
      y2=ins->panning_envelop[right][1];
      x-=x1;
      x2-=x1;
      y2-=y1;
      channels[i].envelop_panning = y1 + x*y2/x2;
      //      channels[i].panning <<=2;
    }
  }

  /* we've done a click, let's inc the speed (which is the number of ticks treated) */
  module->speed++;
  return (44100L*8*5/(module->bpm*2))&(~7); //module->nb_bytes;
}

/* this function will put in the soft mix-buffer exactly one datum, ie 8 bytes (32b mixer + stereo = 4*2 = 8 bytes) */
void handle_one_datum(XM *module)
{
  channel *channels=module->channels;
  int i;
  /* for all channels of the module, see if it's currently in use, then take the datum from the sample,
   * and add it to the mixer-buffer with correct left and right volume (and take care if 8 or 16b sample).
   * Then, increment the current pos in the sample, deal with loop or end of sample. */
  for (i=0; i<module->nb_channels; i++) {
    sample *sh;
    int pos;
    if (!channels[i].in_use)
      continue;
    sh=channels[i].sample;
    pos=channels[i].sample_position>>32;
    /* we may have something wrong, if we change the sample while it's being played */
    /* bad to test it EACH time, only once before to call this fun. Well, this code must all be rewriten... */
    if (!sh || !sh->sample_data || pos>=sh->sample_length) {
      channels[i].in_use=0;
      continue;
    }
    if (sh->sample_type & 0x10) { /* 16b sample */
      short *sample_data;
      sample_data=(short *)sh->sample_data;
      sample_data+=(pos>>1);
      mixer_soft_buffer[mixer_pos]+=((int)*sample_data*channels[i].left_volume)>>6;
      mixer_soft_buffer[mixer_pos+1]+=((int)*sample_data*channels[i].right_volume)>>6;
    } else { /* 8b sample */
      signed char *sample_data;
      sample_data=(signed char *)sh->sample_data;
      sample_data+=pos;
      /* it's -128 .. 127 and we want -32768 .. 32767, so <<8 */
      mixer_soft_buffer[mixer_pos]+=((((int)*sample_data)<<8)*channels[i].left_volume)>>6; 
      mixer_soft_buffer[mixer_pos+1]+=((((int)*sample_data)<<8)*channels[i].right_volume)>>6;
    }
    /* now, let's deal with loop or end */
    channels[i].sample_position+=channels[i].increment;
    pos=channels[i].sample_position>>32;
    if(!(sh->sample_type&3)) {
      /* no loop, so test end of sample, and if yes then channel no more in use */
      if (pos>=sh->sample_length) {
#if 0
	channels[i].in_use=0;
	/* we might have a problem with 2 cut-key played the one after the other. the next reset avoid it. */
	channels[i].instrument=(intrument *)0;
	/* no use, but i prefer to have (so if trouble is there, it will appear one day because of this null-pointer */
	channels[i].sample=(sample *)0;
#endif
	/* we clear all, that's simplier */
	memset(&channels[i], 0, sizeof(channel));
      }
    } else {
      /* let's test loop (ping-pong not yet supported) */
      if (pos>=sh->sample_loop_start+sh->sample_loop_length)
	channels[i].sample_position=(long long)sh->sample_loop_start<<32;
    }
  }
  mixer_pos+=2;
}

/* this function mix at most 1024 bytes or stops at the end of a tick */
int handle_tick(XM *module, int max_handle)
{
  int done=0;

  if (module->to_handle==0)
    module->to_handle=new_line(module);
  if (module->to_handle==0)
    /* if here, we are in dump-mode (we dump to wav), so we don't loop. We then return 0 */
    return 0;

  while(done<max_handle && module->to_handle>0) {
    handle_one_datum(module);
    /* because we have a 32b stereo mixer, so 8 bytes is one datum */
    done+=8;
    module->to_handle-=8;
  }
  return done;
}

void clear_mixer_soft_buffer(void)
{
  memset(mixer_soft_buffer, 0, sizeof(mixer_soft_buffer));
  mixer_pos=0;
}

void translate_mixed_data(void)
{
  int i;
  int *from=(int *)mixer_soft_buffer;
  short *to=(short *)mixer_buffer_soundcard;
  for (i=0; i<1024/2; i++, to++, from++)
    *to=(*from)>>3;
}

/* this function has to write 1024 bytes to the soundcard */
void mix(XM *module)
{
  int nb=1024*2, done;
  clear_mixer_soft_buffer();
  if (module->mix & MODULE_PLAY)
    while(nb) {
      done=handle_tick(module, nb);
      if (done==0) {
	/* if there, we stop the dump, and say how many bytes the last chunk is */
	module->do_we_dump=0;
	module->dumped=(1024*2-nb)>>2;
	return;
      }
      nb-=done;
    }
  /* we don't play anymore when dumping */
  if (module->do_we_dump) {
    module->dumped=512;
    return;
  }
  nb=1024*2;
  mixer_pos=0;
  if (module->mix & LIVE_PLAY)
    while(nb)
      nb-=handle_tick_live(module, nb);
  translate_mixed_data();
  play_sound(mixer_buffer_soundcard, 1024);
}

/* fun called to close all the channels, when we change the current playing module for example */
void close_all_channels(XM *xm)
{
  int i;

  for (i=0; i<32; i++)
    xm->channels[i].in_use=0;
}

/* fun used in sub_interface.interface_subwin_14, when we dec the nb of chan., we must close the "deleted" */
void close_two_channels(int nb, XM *xm)
{
  xm->channels[nb].in_use=xm->channels[nb+1].in_use=0;
}
